package ACL2::Procesador;

#
# Postprocesador de demostraciones de ACL2
# Copyright (C) 2005-2008 Antonio García Domínguez
# $Id$
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use 5.0006;
use warnings;
use strict;

use ACL2::Dependencias;
use ACL2::Localizacion;
use ACL2::Tags;
use ACL2::Orden::Fabrica;
use ACL2::Orden;
use ACL2::Tags;
use ACL2::Lista;
use XML::Writer;
use File::Spec;

our $VERSION = '1.183';
my $lh = ACL2::Localizacion->get_handle();

=head1 FUNCTIONS

Abajo se listan todas las funciones de este mE<oacute>dulo.

=head2 new

  Uso: Crea un nuevo Procesador, sin lanzarlo.
  Responsabilidad: Constructor.
  ParE<aacute>metros: Hash con los siguientes elementos:

=over 2

=item *

lisp: Cadena con el contenido del fichero Lisp sin procesar.

=item *

acl2: Cadena con el contenido de la salida de ACL2 sin procesar.

=item *

fuente_lisp: Cadena con la ruta completa a la fuente Lisp, para posterior referencia.

=item *

deps: Objeto ACL2::Dependencias con las dependencias del fichero actual

=back

  Devuelve: ACL2::Procesador construido.
  Excepciones: ninguna.

=cut

sub new {
    my ($clase, %params) = @_;

    # Nos aseguramos de que la ruta al fuente Lisp sea absoluta
    return bless {
	"lisp" => $params{lisp},
	"acl2" => __PACKAGE__->_procesar_salida_ACL2($params{acl2}),
	"fuente_lisp" => $params{fuente_lisp},
	"fuente_dem"  => $params{fuente_dem},
	"deps" => $params{deps},
    }, $clase;
}

=head2 ejecutar

  Uso: Ejecuta el post-procesador y vuelca su salida sobre writer.
  Responsabilidad: Filtra la salida de ACL2 a XML.
  ParE<aacute>metros: XML::Writer en el que escribir.
  Devuelve: Nada.

=cut

sub ejecutar
{
    my ($this, $writer) = @_;
    
    # Separamos la salida de ACL2 según líneas de prompt
    my @salidas = split /(\S*?) \s !>[^>]/x, $this->{acl2};
    
    # Procesamos la presentación
    $salidas[0]  =~ /ACL2 \s Version \s ([\.0-9]+)/xs 
        or die $lh->maketext("Se esperaba la versión de ACL2");
    my $version  = $1;
    
    $salidas[0]  =~ /Initialized \s with \s (\(.+?\))/xs
       or die $lh->maketext("Se esperaban las opciones de ACL2");
    my $initOpts = $1;
    
    $writer->startTag(ACL2::Tags::MAIN_TAG, 
                      "acl_version"  => $version,
                      "initOpts"     => $initOpts,
		      "source"       => $this->{fuente_lisp},
		      "source_proof" => $this->{fuente_dem});
    
    # Retiramos la bienvenida y despedida
    shift @salidas;
    pop @salidas;

    # Los elementos pares son los paquetes bajo los cuales 
    # están definidas cada una de las órdenes
    my @paquetes = ();
    for (my $i=$#salidas; $i>=0; --$i) {
	unshift(@paquetes, splice(@salidas, $i, 1)) if $i % 2 == 0;
    }

    my @ordenes = __PACKAGE__->analizar_fuente_lisp($this->{lisp});

    # Añadimos la información de dependencias
    $this->{deps}->escribir_xml($writer);

    # Procesamos la salida de cada una de las ordenes
    for (my $i=0; $i <= $#salidas; ++$i) {
	my $orden = ACL2::Orden::Fabrica->crear($ordenes[$i], $salidas[$i], $paquetes[$i]);
	$orden->escribir_xml($writer);
    }
    $writer->endTag(ACL2::Tags::MAIN_TAG);

}

=head2 analizar_fuente_lisp($paquete, $fuenteLisp)

  Uso: ME<eacute>todo estE<aacute>tico pE<uacute>blico.
  Responsabilidades: Realiza el anE<aacute>lisis sintE<aacute>ctico de toda fuente lisp.
  Entrada: 

=over 2

=item *

fuenteLisp - cadena con el cE<oacute>digo fuente Lisp.

=back

  Salida: Lista de cadenas con cada S-expresiE<oacute>n, correspondientes a
  E<oacute>rdenes de ACL2.

=cut

sub analizar_fuente_lisp
{
    my ($paquete, $fuenteLisp) = @_;

    # Retiramos comentarios
    $fuenteLisp =~ s/;.*/ /gm;

    # Obtenemos la lista correspondiente de ordenes
    my @ordenes = ACL2::Lista->dividir_lista("($fuenteLisp)");

    return @ordenes;
}

##################################################
#
# MÉTODOS PRIVADOS
#
##################################################

=head1 _procesar_salida_ACL2

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidades: Procesa la salida de ACL2, simplificando su
    procesamiento sin pE<eacute>rdida de informaciE<oacute>n.
  ParE<aacute>metros: salida - cadena con salida a procesar.
  Devuelve: cadena procesada.

=cut

sub _procesar_salida_ACL2
{
    my ($paquete, $salida) = @_;
    caller eq __PACKAGE__ or die;

    $salida = join '', $salida;

    # Buscamos las regiones que hemos de preservar y limpiamos las demás
    my $nuevaSalida = "";
    my $inicio      = 0;
    my $fin         = 0;
    my $finAnterior = 0;
    while ($salida =~ /
	     ^\([^.]+\n       # Línea inicial sin punto (evitamos predicados cortos)
	        [^\.]+?\)\.$  # Resto de líneas hasta el punto final
	   |
	    (?<=\n)\([^\.]+?\)(?=\n\n) # Párrafo aislado y con una larga S-expresión
	   |
	    ^\({2,} (?:.|\n)*? \){2,}\.$   # Inicio de una lista de listas
	   /gmx) 
    {
	$inicio = $-[0];
	$fin    = $+[0];

	my $prefijo      = substr $salida, $finAnterior, $inicio - $finAnterior;
	my $coincidencia = substr $salida, $inicio, $fin - $inicio;

	# Comprobamos ahora si es realmente un pE<aacute>rrafo de s-expresión
	my @listaLisp = ACL2::Lista->dividir_lista("(${coincidencia})");
	# Añadimos el punto final al último elemento, si es que lo había
	{
	    my $last = pop @listaLisp;
	    if ($last eq ".") {
		push @listaLisp, (pop @listaLisp) . $last;
	    }
	    else {
		push @listaLisp, $last;
	    }
	}

	if (@listaLisp > 1) {
	    # No es una lista, sino que contiene a una
	    foreach (@listaLisp) {
		/^\(.+\n(?:.|\n)+\)\.$/ && do {
		    # Es la lista que estaba contenida
		    $coincidencia = $_;
		    next;
		};

		$prefijo .= " " . $_;
	    }
	}

	$nuevaSalida .= __PACKAGE__->_limpiar_texto_ACL2($prefijo) . "\n\n";
	$nuevaSalida .= $coincidencia . "\n\n";

	$finAnterior = $fin;
    }
    # Añadimos la cola del documento
    $nuevaSalida .= __PACKAGE__->_limpiar_texto_ACL2(substr $salida, $fin);
    
    return $nuevaSalida;
}

=head1 _limpiar_texto_ACL2

  Uso: ME<eacute>todo estE<aacute>tico privado.
  Responsabilidad: 
    Limpia un trozo de texto puro de ACL2, sin S-expresiones
    indentadas en su interior.
  ParE<aacute>metros: cadena con bloque de texto a limpiar.
  Devuelve: cadena con bloque de texto procesado.

=cut

sub _limpiar_texto_ACL2
{
    my ($paquete, $texto) = @_;
    caller eq __PACKAGE__ or die;

    # Simplificamos las RE: simplificamos todos los espacios entre
    # palabras y retiramos el espaciado inicial y final de cada línea.
    $texto =~ s/(?<!\n)\n(?!\n)/ /g;
    $texto =~ s/[ \t\r]+/ /gm;
    $texto =~ s/(?:^ +| +$)//gm;

    # Arreglamos nombres de regla esparcidos por saltos de línea tras
    # alguno de sus guiones
    $texto =~ s/-\s+(?=\S)/-/g;

    # Retiramos los guiones en el caso de que vengan tras un indicador de paquete
    $texto =~ s/::-/::/g;
    
    # Retiramos barras inversas (con espacio en blanco asociado)
    $texto =~ s/\\\s+//g;

    return $texto;
}

=head1 NAME

ACL2::Procesador - ME<oacute>dulo para post-procesado de demostraciones de ACL2

=head1 VERSION

Version 1.01

=head1 SYNOPSIS

Transforma un guiE<oacute>n LISP y su demostraciE<oacute>n ACL2 a XML fE<aacute>cil de
procesar por un programa externo.

  use XML::Writer;
  use ACL2::Procesador;
  my $w = new XML::Writer();
  my $p = new ACL2::Procesador("lisp" => $guionLisp,
	     	               "acl2" => $salidaACL2,
			       "fuente_lisp" => "/tmp/demostracion.out");
  $w->xmlDecl();
  $w->doctype(ACL2::Tags::MAIN_TAG, undef, DTD_FILE_FILENAME);  
  $p->ejecutar($w);
  $w->end();

=head1 AUTHOR

Antonio GarcE<iacute>a-DomE<iacute>nguez, C<< <nyoescape at gmail.com> >>

=head1 BUGS

Please report any bugs or feature requests to
C<bug-acl2-procesador at rt.cpan.org>, or through the web interface at
L<http://rt.cpan.org/NoAuth/ReportBug.html?Queue=ACL2-Procesador>.
I will be notified, and then you'll automatically be notified of progress on
your bug as I make changes.

=head1 SUPPORT

You can find documentation for this module with the perldoc command.

    perldoc ACL2::Procesador

=head1 ACKNOWLEDGEMENTS

I'd like to thank my end of degree project supervisors, Francisco
Palomo Lozano and Inmaculada Medina Bulo, for their advice and
support.

=head1 COPYRIGHT & LICENSE

Copyright (C) 2005-2008 Antonio GarcE<iacute>a-DomE<iacute>nguez

This program is free software; you can redistribute it 
and/or modify it under the terms of the GNU General Public 
License as published by the Free Software Foundation; 
either version 2 of the License, or (at your option) any 
later version.

This program is distributed in the hope that it will be 
useful, but WITHOUT ANY WARRANTY; without even the implied 
warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR 
PURPOSE. See the GNU General Public License for more 
details.

You should have received a copy of the GNU General Public 
License along with this program; if not, write to the Free 
Software Foundation, Inc., 59 Temple Place, Suite 330, 
Boston, MA 02111-1307 USA

=cut

1; # End of ACL2::Procesador
__END__
