package ACL2::Resultado;

=head1 NAME

ACL2::Resultado - resultado de una demostraciE<oacute>n de ACL2

=cut

#
#    Módulo de procesado de demostraciones ACL2
#    Copyright (C) 2006-2008 Antonio García Domínguez
#
#    This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
#

use strict;
use warnings;
use ACL2::Tags;
use ACL2::Localizacion;
use ACL2::Orden;
use ACL2::Meta;
use ACL2::Proceso;

my $lh = ACL2::Localizacion->get_handle();

=head1 FUNCTIONS

Las funciones que aporta este mE<oacute>dulo son:

=head2 new

Uso: Constructor.

Responsabilidad: Crea un nuevo objeto ACL2::Resultado, sin lanzarlo.

ParE<aacute>metros: 

=over 2

=item *

formula - S-expresiE<oacute>n a demostrar.

=item *

salida - salida de ACL2 para esta fE<oacute>rmula.

=back

=cut

sub new
{
    my ($clase, $formula, $salida) = @_;

    return bless {
	"formula" => $formula,
	"salida" => $salida,
    }, $clase;
}

## Accesores automáticos usando clausuras lambda

=head2 get_formula

Devuelve cadena con la S-expresiE<oacute>n de la orden.

=head2 get_long_dem

Devuelve longitud en caracteres de la respuesta a la orden.

=head2 get_salida

Devuelve cadena con la respuesta de ACL2 a la orden.

=head2 set_formula($nuevaFormula)

Cambia la cadena con la S-expresiE<oacute>n usada en la orden.

=head2 set_long_dem($nuevaLong)

Cambia la longitud en caracteres de la respuesta a la orden.

=head2 set_salida($nuevaSalida)

Cambia la cadena con la respuesta de ACL2 a la orden.

=cut

{
    no strict 'refs';

    for my $campo (qw(salida formula long_dem)) {
	*{"get_$campo"} = sub {
	    my ($this) = @_;
	    return $this->{$campo};
	};
	*{"set_$campo"} = sub {
	    my ($this, $nuevo) = @_;
	    my $anterior = $this->{$campo};
	    $this->{$campo} = $nuevo;
	    return $anterior;
	};
    }
}


=head2 demostrar

Uso: ME<eacute>todo pE<uacute>blico.

Responsabilidad: Decide si ha tenido E<eacute>xito la
demostraciE<oacute>n. Da el valor correcto a long_dem con la longitud
de la demostraciE<oacute>n.

ParE<aacute>metros: Ninguno.

Devuelve: nada.

=cut    

sub demostrar
{
    my ($this) = @_;

    my $salida_orig = $this->get_salida;

    ## Toda demostración tiene tres partes: su enunciado, el
    ## resultado final (éxito/fracaso) y un sumario.
    my ($texto, $resultado) = split /
	(Consequently, \s the \s proof 
	 \s attempt \s has \s failed\. # Fracaso en general
	 | Obviously, \s the \s proof
	 \s attempt \s has \s failed\. # Fracaso en general
	 | \QQ.E.D.\E)		 # Éxito en general
	/x, $this->get_salida;

    $this->{exito} = ($resultado =~ /\QQ.E.D.\E/);
    $this->{long_dem} = length($texto) + length($resultado);

    # Marcamos la meta inicial y las inducciones
    $this->set_salida("Goal\n" . $this->{formula} . ".\n\n" . $texto);
    $this->_obtener_formulas_induccion;
    $this->_marcar_inducciones;

    # Dividimos en metas (no existe prefijo, pero split devolverá algo de todas formas)
    my (undef, @salida_metas) = split /(^Subgoal \S*|^Goal\S*)/m, $this->get_salida;
  
    # Creamos las metas, inicializándose estas al crearse
    $this->{metas}  = {};
    for (my $i = 0; $i < @salida_metas; ) {
	my $etiq   = $salida_metas[$i++];
	my $salida = $salida_metas[$i++];
	$this->{metas}->{$etiq} = new ACL2::Meta($etiq, $salida);
    }
    # Asociamos cada meta con su meta padre (Goal es la raíz)
    for (my $i = 0; $i < @salida_metas; $i += 2) {
	my $etiq      = $salida_metas[$i];
	my $etiqPadre = $this->_obtener_meta_padre($etiq);

	if (!defined($etiqPadre)) { next; }
	$this->{metas}->{$etiqPadre}->agregar_submeta($this->{metas}->{$etiq});
    }
    
    return $this->{exito};
}


=head2 escribir_xml

Uso: ME<eacute>todo pE<uacute>blico.

Responsabilidades: Escribe la informaciE<oacute>n de la
demostraciE<oacute>n en formato XML.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub escribir_xml
{
    my ($this, $writer) = @_;
    $this->{metas}->{Goal}->escribir_xml($writer);
}


#
# MÉTODOS PRIVADOS
#

=head2 _obtener_formulas_induccion

Uso: ME<eacute>todo privado.

Responsabilidad: Obtiene la informaciE<oacute>n de las
fE<oacute>rmulas demostrados por inducciE<oacute>n y sus etiquetas,
almacenE<aacute>ndola en la instancia.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut

sub _obtener_formulas_induccion
{
    caller eq __PACKAGE__ or die;
    my ($this) = @_;

    # Diccionario de fórmulas de inducción
    my %formulas_ind;

    my @elems = 
	$this->{salida} =~ /
	(?-x:(\([^.]*\))\.)    # Fórmula a etiquetar
	\s+
	(?-x:Name the formula above (.*?)\.\s) # Etiqueta
	/oxmg;

    while (@elems > 0) {
	my ($formula, $etiq) = splice @elems, 0, 2;
	$formulas_ind{$etiq} = $formula;
    }

    $this->{formulas_ind} = \%formulas_ind;
}



=head2 _marcar_inducciones

Uso: ME<eacute>todo privado.

Responsabilidad: Marca las inducciones con etiquetas de la forma
Subgoal *N, aE<ntilde>adiendo tambiE<eacute>n sus fE<oacute>rmulas.

ParE<aacute>metros: Ninguno.

Devuelve: Nada.

=cut    

sub _marcar_inducciones
{
    caller eq __PACKAGE__ or die;
    my ($this) = @_;

    my $rh_form_ind    = $this->{formulas_ind};
    my $anteriorSalida = $this->get_salida;
    my $nuevaSalida;
    my $prefijo = 0;
    my $inicio  = 0;
    my $final;

    while ($anteriorSalida =~ /
	 # Puede aparecer más texto antes de "Perhaps...", como el SGC
	 (?-x:So we now return to (.*?), which is)        # 1
	   .*?
	 (?-x:Perhaps we can prove \1 by induction\.) 
	 |
	 (?-x:So we now return to (.*?), which is) # 2
	   .*?
	 (?-x:No induction schemes are suggested by \2\.)      
	 |
	 (?-x:We therefore turn our attention to (.*?), which is) # 3
	   .*?
	 (?-x:Perhaps we can prove \3 by induction\.) 
	 |
	 (?-x:We therefore turn our attention to (.*?), which is) # 4
	   .*?
	 (?-x:No induction schemes are suggested by \4\.)
	 |
	 (?-x:So we now return to (.*?), which is) # 5
	 |
	 (?-x:We therefore turn our attention to (.*?), which is) # 6
	 |
	 (?-x:Perhaps we can prove (.*?) by induction\.) # 7
	 |
	 (?-x:No induction schemes are suggested by (.*?)\.) # 8
	/oxsg) 
    {
	my $etiq = $1 || $2 || $3 || $4 || $5 || $6 || $7 || $8;

	$inicio = $-[0];
	$final  = $+[0];

	if (!exists $rh_form_ind->{$etiq}) {
	    die $lh->maketext("No se pudo obtener la fórmula de la etiqueta [_1]", $etiq);
	}

	# Añadimos el prefijo al texto
	$nuevaSalida .= substr $anteriorSalida, $prefijo, $inicio - $prefijo;
	# Etiqueta y fórmula
	$nuevaSalida .= "Subgoal $etiq\n" . $rh_form_ind->{$etiq} . ".\n\n";
	# Texto de la inducción
	$nuevaSalida .= substr $anteriorSalida, $inicio, $final - $inicio;

	# El final de esta meta marca el inicio del prefijo de la siguiente
	$prefijo = $final;
    }
    # Añadimos el texto posterior a la última inducción
    $nuevaSalida .= substr $anteriorSalida, $prefijo;

    # Fijamos el nuevo texto
    $this->set_salida($nuevaSalida);
}


=head2 _obtener_meta_padre

Uso: ME<eacute>todo privado.

Responsabilidad: calcula la etiqueta de la meta padre.

ParE<aacute>metros: etiq - cadena con la etiqueta de la meta hija.

Devuelve: cadena con la etiqueta de la meta padre, o undef si es la
raE<iacute>z.

=cut

sub _obtener_meta_padre
{
    caller eq __PACKAGE__ or die;
    my ($this, $etiq) = @_;

    my $etiqPadre;
    for ($etiq) {
        # Submeta de otra submeta (Goal'' o Subgoal *1/2'3')
	/(.*?)(?:\'([0-9]+)\'$|\'$)/ && do {
	    if (!defined($2)) {
		# La expresión regular ya ha quitado la comilla extra
		$etiqPadre = $1;
	    } elsif ($2 - 1 > 3) {
		$etiqPadre = $1 . "'" . ($2 - 1) . "'";
	    } else {
		$etiqPadre = $1 . ("'" x ($2 - 1));
	    }
	    last;
	};

        # Símbolo no tautológico dentro de inducción (Subgoal *1/1)
        m|(.*)/[0-9]+$| && do {
	    $etiqPadre = $1;
	    last;
	};

	# Submeta resultado de división de la última de Goal, Goal'... en varias
        /Subgoal [0-9]+$/ && do {
	    # Buscamos la meta derivada de Goal más interna
	    my $maxprof   = 0;
	    $etiqPadre = "Goal";
	    foreach (keys %{$this->{metas}}) {
		/Goal(?:\'([0-9]+)\'|(\'+))/ or next;

		my $prof = $1 || length $2;
		if ($prof > $maxprof) {
		    $maxprof   = $prof;
		    $etiqPadre = $_;
		}
	    }

	    last;
	};

        # Submeta de inducción (Subgoal *1.1, Subgoal 3)
        /Subgoal \*[.0-9]+$/ && do {
	    $etiqPadre = "Goal";
	    last;
	};

	# Meta raíz
        /Goal/ && do {
	    $etiqPadre = undef;
	    last;
	};

        die $lh->maketext("No se reconoce la etiqueta [_1]", $etiq);
    }

    return $etiqPadre;
}

1;
